@@ -1,17 +1,20 @@
-import re
-import glob
-import os
-from os.path import join, expanduser
-from gettext import gettext as _
+from deskbar.core.Utils import get_xdg_data_dirs, is_program_in_path, spawn_async, is_executable, PATH
 from deskbar.defs import VERSION
+from deskbar.handlers.actions.OpenDesktopFileAction import OpenDesktopFileAction, parse_desktop_file, parse_desktop_filename
+from deskbar.handlers.actions.OpenWithApplicationAction import OpenWithApplicationAction
+from gettext import gettext as _
+from os.path import join, expanduser, isdir
+import deskbar, deskbar.core.Indexer
+import deskbar.core.Utils
+import gnomedesktop
+import deskbar.interfaces.Action
+import deskbar.interfaces.Match
+import deskbar.interfaces.Module
+import glob
 import gobject
 import gtk
-import deskbar, deskbar.core.Indexer, deskbar.core.Utils
-import deskbar.interfaces.Module, deskbar.interfaces.Match, deskbar.core.gnomedesktop
-from deskbar.core.Utils import get_xdg_data_dirs, is_program_in_path, spawn_async
-from deskbar.handlers.actions.OpenWithApplicationAction import OpenWithApplicationAction
-from deskbar.handlers.actions.OpenDesktopFileAction import OpenDesktopFileAction, parse_desktop_file, parse_desktop_filename
-import deskbar.interfaces.Action
+import os
+import re
 
 HANDLERS = [
     "ProgramsHandler",
@@ -21,6 +24,7 @@ HANDLERS = [
 
 EXACT_MATCH_PRIO = 50
 EXACT_WORD_PRIO = 5
+DESKTOP_FILE_PRIO = 25
 
 class GenericAction(OpenWithApplicationAction):
     
@@ -69,6 +73,9 @@ class GenericProgramMatch(deskbar.interfaces.Match):
 
     def get_hash(self):
         return "generic_"+self._display_prog
+    
+    def get_program(self):
+        return self._display_prog
         
 class GnomeDictMatch(GenericProgramMatch):
     def __init__(self, **args):
@@ -130,8 +137,8 @@ class GnomeDictHandler(SpecialProgramHandler):
     
     def create_match(self, desktop, f):
         return GnomeDictMatch(
-                    name=desktop.get_localestring(deskbar.core.gnomedesktop.KEY_NAME),
-                    icon=desktop.get_string(deskbar.core.gnomedesktop.KEY_ICON),
+                    name=desktop.get_localestring(gnomedesktop.KEY_NAME),
+                    icon=desktop.get_string(gnomedesktop.KEY_ICON),
                     desktop=desktop,
                     desktop_file=f)
         
@@ -154,8 +161,8 @@ class GnomeSearchHandler(SpecialProgramHandler):
     
     def create_match(self, desktop, f):
         return GnomeSearchMatch(
-                    name=desktop.get_localestring(deskbar.core.gnomedesktop.KEY_NAME),
-                    icon=desktop.get_string(deskbar.core.gnomedesktop.KEY_ICON),
+                    name=desktop.get_localestring(gnomedesktop.KEY_NAME),
+                    icon=desktop.get_string(gnomedesktop.KEY_ICON),
                     desktop=desktop,
                     desktop_file=f)
         
@@ -181,8 +188,8 @@ class DevhelpHandler(SpecialProgramHandler):
     
     def create_match(self, desktop, f):
         return DevhelpMatch(
-                    name=desktop.get_localestring(deskbar.core.gnomedesktop.KEY_NAME),
-                    icon=desktop.get_string(deskbar.core.gnomedesktop.KEY_ICON),
+                    name=desktop.get_localestring(gnomedesktop.KEY_NAME),
+                    icon=desktop.get_string(gnomedesktop.KEY_ICON),
                     desktop=desktop,
                     desktop_file=f)
         
@@ -208,12 +215,12 @@ class OpenPathProgramAction(deskbar.interfaces.Action):
                 import subprocess
 
                 prog = subprocess.Popen(
-                    text.split(" "),
+                    self._name.split(" "),
                     stdout=subprocess.PIPE,
                     stderr=subprocess.STDOUT)
                 
                 zenity = subprocess.Popen(
-                    ["zenity", "--title="+text,
+                    ["zenity", "--title="+self._name,
                         "--window-icon="+join(deskbar.ART_DATA_DIR, "generic.png"),
                         "--width=700",
                         "--height=500",
@@ -228,7 +235,7 @@ class OpenPathProgramAction(deskbar.interfaces.Action):
                 #No zenity, get out of the if, and launch without GUI
                 pass
         
-        spawn_async(text.split(" "))
+        spawn_async(self._name.split(" "))
         
     def get_hash(self):
         if self.use_terminal:
@@ -238,9 +245,9 @@ class OpenPathProgramAction(deskbar.interfaces.Action):
 
     def get_verb(self):
         if self.use_terminal:
-            return _("Execute %s in terminal") % "<b>%(text)s</b>"
+            return _("Execute %s in terminal") % "<b>%(name)s</b>"
         else:
-            return _("Execute %s") % "<b>%(text)s</b>"
+            return _("Execute %s") % "<b>%(name)s</b>"
 
 class PathProgramMatch(deskbar.interfaces.Match):
     
@@ -252,6 +259,17 @@ class PathProgramMatch(deskbar.interfaces.Match):
         
     def get_hash(self):
         return self._name
+   
+class StartsWithPathProgramMatch(deskbar.interfaces.Match):
+    
+    def __init__(self, command, priority=0, **args):
+        deskbar.interfaces.Match.__init__(self, name=command, icon="gtk-execute", category="actions", **args)
+        self._command = command
+        self.add_action( OpenPathProgramAction(command, False), True )
+        self.add_action( OpenPathProgramAction(command, True) )
+        
+    def get_hash(self):
+        return self._command
         
 class ProgramsHandler(deskbar.interfaces.Module):
     
@@ -268,22 +286,51 @@ class ProgramsHandler(deskbar.interfaces.Module):
         self._scan_desktop_files()
         
     def query(self, query):
-        result = self.query_path_programs(query)
-        result += self.query_desktop_programs(query)
-        self.set_priority_for_matches(result)
+        result = self.query_desktop_programs(query)
+        
+        desktop_progs = set()
+        for match in result:
+            desktop_progs.add( match.get_program () )
+            match.set_priority (self.get_priority() + DESKTOP_FILE_PRIO + match.get_priority())
+        
+        path_result = self.query_path_programs(query, desktop_progs)
+        result += path_result
+        
         self._emit_query_ready(query, result )
         
-    def query_path_programs(self, query):
-        program = query.split(" ")[0]
-        if is_program_in_path(program):
-            return [PathProgramMatch(program, query)]
+    def query_path_programs(self, query, desktop_progs):
+        """
+        @param query: Query string
+        @param desktop_progs: Names of binaries from .desktop files
+        """
+        args = query.split(" ")     
+        program = args[0]
+        
+        if len(args) == 1:
+            results = []
+            for pathdir in PATH:
+                if isdir(pathdir) and is_executable(pathdir):
+                    for f in os.listdir(pathdir):
+                        pathprog = join(pathdir, f)
+                        if (not (f in desktop_progs)) and not isdir(pathprog) \
+                        and f.lower().startswith(program) and is_executable(pathprog):
+                            match = StartsWithPathProgramMatch(f)
+                            match.set_priority (self.get_priority() + get_priority_for_name(query, f))
+                            results.append( match )
+            return results
         else:
-            return []
+            # We have arguments, execute the command as typed in by the user
+            if not (program in desktop_progs) and is_program_in_path(program):
+                match = PathProgramMatch(program, query)
+                match.set_priority (self.get_priority() + EXACT_MATCH_PRIO)
+                return [match]
+            else:
+                return []
     
     def query_desktop_programs(self, query):
         result = []
         for match in self._indexer.look_up(query):
-            match._priority = get_priority_for_name(query, match._desktop.get_string("Exec"))
+            match.set_priority (get_priority_for_name(query, match._desktop.get_string("Exec")))
             result.append(match)
         return result
                 
@@ -294,20 +341,21 @@ class ProgramsHandler(deskbar.interfaces.Module):
                     result = parse_desktop_file(f)
                     if result != None:
                         match = GenericProgramMatch(
-                                    name=result.get_localestring(deskbar.core.gnomedesktop.KEY_NAME),
-                                    icon=result.get_string(deskbar.core.gnomedesktop.KEY_ICON),
+                                    name=result.get_localestring(gnomedesktop.KEY_NAME),
+                                    icon=result.get_string(gnomedesktop.KEY_ICON),
                                     desktop=result,
                                     desktop_file=f)
                         self._indexer.add("%s %s %s %s %s" % (
                                     result.get_string("Exec"),
-                                    result.get_localestring(deskbar.core.gnomedesktop.KEY_NAME),
-                                    result.get_localestring(deskbar.core.gnomedesktop.KEY_COMMENT),
-                                    result.get_string(deskbar.core.gnomedesktop.KEY_NAME),
-                                    result.get_string(deskbar.core.gnomedesktop.KEY_COMMENT),
+                                    result.get_localestring(gnomedesktop.KEY_NAME),
+                                    result.get_localestring(gnomedesktop.KEY_COMMENT),
+                                    result.get_string(gnomedesktop.KEY_NAME),
+                                    result.get_string(gnomedesktop.KEY_COMMENT),
                                 ), match)
 
 def get_priority_for_name(query, name):
-    if name.split(" ")[0].endswith(query):
+    bin = name.split(" ")[0]
+    if bin == query:
         return EXACT_MATCH_PRIO
     else:
         return EXACT_WORD_PRIO
